.de P2
..
.TL
La organización de las redes en Plan 9
.AU
Dave Presotto
Phil Winterbottom
.sp
presotto,philw@plan9.bell-labs.com
.AB
.FS
Originalmente aparecido en 
.I
Proc. of the Winter 1993 USENIX Conf.,
.R
pp. 271-280,
San Diego, CA
.FE
En un sistema distribuido las redes son de importancia capital. Este documento describe la implementación, filosofía de diseño y organización del soporte de red en Plan 9. Hace referencia a los requerimientos de red para los sistemas distribuidos, nuestra implementación del nucleo, los nombres de red, los interfaces de usuario y el rendimiento. También observamos que gran parte de esta organización es importante en los sistemas actuales.
.AE
.NH
Introducción
.PP
Plan 9 [Pike90] es un sistema portable distribuido, multi usuario, de propósito general implementado en una gran variedad de ordenadores y redes.
Lo que lo diferencia es su organización. El objetivo de esta organización fué reducir la administración y promover la compartición de recursos. Una de las claves de su éxito como sistema distribuido es la organización y gestión de sus redes.
.PP
Un sistema Plan 9 comprende servidores de ficheros, servidores de CPU y terminales. Los servidores de ficheros y de CPU son normalmente máquinas multiprocesador centrales con amplia memoria y conectados con redes de alta velocidad.
Una variedad de máquinas tipo estación de trabajo sirven como terminales de los servidores centrales usando varias redes y protocolos.
La arquitectura del sistema exige una jerarquía de velocidades de red acordes con las necesidades de sus componentes.
La conexión entre servidores de ficheros y servidores de CPU suele consistir en enlaces punto a punto de fibra óptica de banda ancha.
Las conexiones entre servidores y terminales suelen ser de velocidad media, como Ethernet [Met80] y Datakit [Fra80].
Conexiones de baja velocidad por Internet y el backbone de AT&T sirve para usuarios de Oregon e Illinois.
Lineas básicas de datos RDSI y lineas serie a 9600 baudios proporcionan acceso a los usuarios en casa.
.PP
Puesto que los servidores de CPU y los terminales usan el mismo nucleo, los usuario pueden elegir ejecutar programas localmente en sus terminales o remotamente en los servidores de CPU.
La organización de Plan 9 oculta los detalles de la conectividad del sistema permitiendo tanto a usuarios como a administradores configurar su entorno para que sea tan distribuido o centralizado como ellos deseen.
Mediante comandos simples se puede construir un espacio de nombres local que abarque muchas máquinas y redes.
En el trabajo, los usuarios tienden a usar sus terminales como estaciones de trabajo, ejecutando programas interactivos localmente y reservando los servidores de CPU para datos o trabajos de computación intensiva, como compilar programas o ejecutar programas de ajedrez.
En casa, cuando se conectan a través de redes lentas, tienden a realizar más trabajo en los servidores de CPU para minimizar el tráfico por la linea.
El objetivo de la organización de red es proporcionar el mismo entorno al usuario estén donde estén los recursos que utilice.
.NH 
Soporte de red del nucleo
.PP
Las redes juegan un papel importante en un sistema distribuido. Esto es particularmente cierto en Plan 9, donde la mayoría de los recursos son proporcionados por servidores externos al nucleo.
La importancia del código de red dentro del nucleo se refleja en su tamaño: de sus 25.000 lineas de código, 12.500 son relativas a red y a protocolos. Constantemente se están agregando redes y la fracción de código de comunicaciones está creciendo.
Peor aún, el código de red es complejo. Las implementaciones de protocolos consisten casi por entero en sincronización y manejo dinámico de memoria, areas que demandan sutiles estrategias de recuperación de errores.
El nucleo actual soporta Datakit, enlaces de fibra óptica punto a punto, una suite de protocolos de Internet (IP) y un servicio de datos RDSI.
La variedad de redes y máquinas ha hecho surgir cuestiones no planteadas por otros sistemas que se ejecutan en hardware comercial y que solamente soportan Ethernet o FDDI.
.NH 2
El protocolo de sistema de ficheros
.PP
Una idea central de Plan 9 es la representación de un recurso como un sistema de ficheros jerárquico.
Cada proceso ensambla una vista del sistema construyendo un 
.I "espacio de nombres
[Needham] que conecta sus recursos.
Los sistemas de ficheros no necesitan representar ficheros en disco; de hecho, la mayoría de los sistemas de ficheros de Plan 9 no tienen almacenamiento permanente.
Un sistema de ficheros típico representa dinámicamente algún recurso como un conjunto de conexiones de red o la tabla de procesos. La comunicación entre el nucleo, los drivers de dispositivos y los servidores de ficheros remotos o locales usa un protocolo llamado 9P. Dicho protocolo consiste en 17 tipos de mensajes, que describen operaciones sobre ficheros y directorios.
Los dispositivos residentes del nucleoy los drivers de protocolos usan una versión procedural del protocolo, mientras que los servidores de ficheros externos usan una forma RPC.
Casi todo el tráfico entre sistemas Plan 9 consiste en mensajes 9P. Este protocolo se vasa en varias propiedades del protoloco de transporte subyacente. Asume que los mensajes llegan de manera fiable, y en correcta secuencia, y que se preservan los delimitadores entre mensajes.
Cuando un protocolo no cumple estos requerimientos (por ejemplo, TCP no preserva los delimitadores) nosotros proporcionamos mecanismos para reorganizar los mensajes antes de dejar que el sistema los maneje.
.PP
Una estructura de datos del nucleo, el 
.I canal
es un handle a un servidor de ficheros.
Las operaciones sobre un canal generan los mensajes 9P siguientes:
Los mensajes 
.CW session
y
.CW attach
autentican la conexión, establecida por medios externos a 9P, y validan a su usuario.
El resultado es un canal autenticado que hace referencia a la raíz del servidor.
El mensaje
.CW clone
crea un nuevo canal idéntico a otro existente, de forma similar a como hace la llamada del sistema
.CW dup.
Un canal puede ser movido a un fichero del sistema usando el mensaje 
.CW walk
para descender cada nivel en la jerarquía.
Los mensajes
.CW stat
y
.CW wstat
leen y escriben los atributos del fichero referenciado por el canal.
El mensaje
.CW open
prepara un canal para una posterior operación mediante los mensajes
.CW read
y
.CW write
que acceden a los contenidos del fichero.
.CW Create
y
.CW remove
realizan las acciones que sus nombres indican sobre el fichero
El mensaje
.CW clunk
descarta un canal sin afectar al fichero.
.PP
Un servidor de ficheros residente del nucleo llamado
.I "driver mount "
convierte la versión procedural de 9P en RPCs.
La llamada del sistema
.I mount
proporciona un descriptor de fichero, que puede ser un pipe a un proceso de usuario o a una conexión de red con una máquina remota, para ser asociado con el punto de montaje.
Despues de un mount, las operaciones en el arbol de ficheros bajo el punto de montaje son enviadas como mensajes al servidor de ficheros.
El driver mount maneja buffers, empaqueta y desempaqueta parámetros de los mensajes, y desmultiplexa entre procesos que usan el servidor de ficheros.
.NH 2
Organización del nucleo
.PP
El código de red en el nucleo está dividido en tres capas: interfaz con el hardware, procesamiento de protocolos e interfaz con programas.
Un driver de dispositivo típicamente usa flujos para conectar las dos capas de interfaz. Modulos de flujos adicionales pueden ser incluidos en un dispositivo para procesar protocolos. Cada driver de dispositivo es un sistema de ficheros residente del nucleo que contiene solamente unos pocos ficheros: por ejemplo, representamos cada UART por medio de un fichero control y otro data.
.P1
cpu% cd /dev
cpu% ls -l eia*
--rw-rw-rw- t 0 bootes bootes 0 Jul 16 17:28 eia1
--rw-rw-rw- t 0 bootes bootes 0 Jul 16 17:28 eia1ctl
--rw-rw-rw- t 0 bootes bootes 0 Jul 16 17:28 eia2
--rw-rw-rw- t 0 bootes bootes 0 Jul 16 17:28 eia2ctl
cpu%
.P2
El fichero control se usa para controlar el dispositivo; escribiendo la cadena
.CW b1200
en
.CW /dev/eia1ctl
set pone la linea a 1200 baudios.
.PP
Los dispositivos multiplexados presentan un interfaz de estructura un poco más compleja. Por ejemplo, el driver Ethernet LANCE sirve un árbol de ficheros de dos niveles (Figura 1) proporcionando
.IP \(bu
control y configuración de dispositivo
.IP \(bu
protocolos de nivel usuario como ARP
.IP \(bu
interfaces de diagnóstico para software de snooping
.LP
El directorio superior contiene un fichero
.CW clone
y un directorio para cada conexión, numerados desde
.CW 1
a
.CW n .
Cada directorio de conexión corresponde a un tipo de paquete Ethernet.

??????
Opening the
.CW clone
file finds an unused connection directory and opens its
.CW ctl.
file.

Abrir el fichero 
.CW clone
hace que se encuentre un directorio de una conexión no utilizada y que se abra su fichero
.CW ctl.
?????????

Leyendo el fichero control se obtiene el numero de la conexión en ASCII; el proceso usuario puede usar este valor para construir el nombre del directorio de la conexión.
En cada directorio de conexión, ficheros llamados
.CW ctl , 
.CW data , 
.CW stats ,
y
.CW type
proporcionan acceso a la conexión.
Escribir la cadena
.CW "connect 2048"
en el fichero 
.CW ctl
hace que se ponga el tipo de paquete a 2048 y configura la conexión para recibir todos los paquetes IP enviados a la máquina. Las posteriores lecturas del fichero 
.CW type
obtendrán la cadena
.CW 2048 .
El fichero
.CW data
accede al medio: leyéndolo se obtiene el siguiente paquete del tipo seleccionado. Escribiendo en él se pone se agrega a la cola un paquete para su transmisión, después de agregar una cabecera que contiene la dirección de origen y el tipo de paquete.
El fichero
.CW stats
devuelve un texto ASCII que contiene la dirección del interfaz, la cuenta de paquetes de entrada y de salida, las estadísticas de error, e información general sobre el estado del interfaz.
??????????
?????.so tree.pout
???????????
.PP
Si varias conexiones en un interfaz están configuradas para un tipo concreto de paquete, cada una recibe una copia de los paquetes entrantes. El tipo especial de paquete
.CW -1
selecciona todos los paquetes.
Si se escriben las cadenas
.CW promiscuous
y
.CW connect
.CW -1
en el fichero
.CW ctl
se configura una conversación para recibir todos los paquetes del Ethernet.
.PP
Aunque el interfaz del driver pueda parecer muy elaborado, la representación de un dispositivo como un conjunto de ficheros que usan cadenas ASCII para comunicarse tiene varias ventajas. Cualquier mecanismo que soporte acceso remoto a ficheros permite inmediatamente a una máquina remota utilizar nuestros interfaces como pasarela.
Usar cadenas ASCII para controlar el interfaz evita problemas de orden de byte y asegura una representación uniforme para dispositivos en la misma máquina e incluso permite que los dispositivos sean accedidos remotamente. Representar dispositivos diferentes con el mismo conjunto de ficheros permite que herramientas comunes sirvan para varias redes o interfaces.
Programas como 
.CW stty
son reemplazados por
.CW echo
y una redirección de la shell
.NH 2
Dispositivos de protocolo
.PP
Las conexiones de red se representan mediante pseudo-dispositivos llamados dispositivos de protocolo.
Hay drivers para dispositivos de protocolo para el protocolo URP Datakit y para cada uno de los protocolos de Internet, TCP, UDP e IL.
IL, que describimos más adelante, es un nuevo protocolo de comunicación usado por Plan 9 para transmitir RPC's de sistemas de ficheros.
Todos los dispositivos de protocolo parecen idénticos, así que los programas usuarios no contienen código específico de red.
.PP
Cada driver de dispositivo de protocolo sirve una estructura de directorios similar a la del driver de Ethernet.
El directorio superior contiene un fichero
.CW clone
y un directorio para cada conexión numeradas de
.CW 0
a
.CW n .
Cada directorio de conexión contiene ficheros para controlar una conexión y para enviar y recibir información.
Un directorio de conexión TCP tiene el siguiente aspecto:
.P1
cpu% cd /net/tcp/2
cpu% ls -l
--rw-rw---- I 0 ehg    bootes 0 Jul 13 21:14 ctl
--rw-rw---- I 0 ehg    bootes 0 Jul 13 21:14 data
--rw-rw---- I 0 ehg    bootes 0 Jul 13 21:14 listen
--r--r--r-- I 0 bootes bootes 0 Jul 13 21:14 local
--r--r--r-- I 0 bootes bootes 0 Jul 13 21:14 remote
--r--r--r-- I 0 bootes bootes 0 Jul 13 21:14 status
cpu% cat local remote status
135.104.9.31 5012
135.104.53.11 564
tcp/2 1 Established connect
cpu%
.P2
Los ficheros
.CW local ,
.CW remote ,
y
.CW status
proporcionan información sobre el estado de la conexión.
Los ficheros 
.CW data
y
.CW ctl
???????????
proporcionan acceso a la parte de proceso del flujo que implementa el protocolo.
provide access to the process end of the stream implementing the protocol.
???????????

El fichero
.CW listen
se usa para aceptar llamadas entrantes desde la red.
.PP
Para establecer una conexión se siguen estos pasos:
.IP 1)
El dispositivo clone del protocolo apropiado se abre para reservar una conexión no usada.
.IP 2)
El descriptor de fichero devuelto por la orden open apunta al fichero
.CW ctl
de la nueva conexión.
Leyendo este fichero se obtiene una cadena ASCII que contiene el número de la conexión.
.IP 3)
Una cadena ASCII específica de la red y del protocolo se escribe en el fichero
.CW ctl.
.IP 4)
El path del fichero
.CW data
se construye usando el número de conexión.
Cuando el fichero
.CW data
se abre, la conexión está establecida.
.LP
Un proceso puede leer y escribir este descriptor de fichero para enviar y recibir mensajes en la red.
Si el proceso abre el fichero
.CW listen
se bloquea hasta que reciba una llamada entrante.
Una cadena de dirección escrita en 
.CW ctl
antes de abrir listen selecciona el puerto o servicio que el proceso está listo para aceptar.
Cuando se recibe una llamada, la apertura se completa y se devuelve un descriptor de fichero apuntando al fichero 
.CW ctl
de la nueva conexión.
Leer del fichero
.CW ctl
proporciona un número de conexión que se usa para construir el path del fichero
.CW data.
Una conexión permanece establecida mientras cualquiera de los ficheros del directorio de la misma están referenciados, o hasta que se recibe un close desde la red.
.NH 2
Corrientes
.PP
Una
.I stream 
o corriente [Rit84a][Presotto] es un canal bidireccional que conecta un dispositivo (físico o no) con un proceso de usuario.El proceso de usuario inserta y retira datos en un lado de la corriente. Los procesos del nucleo que actúan en nombre de un dispositivo insertan datos al otro lado.
Los canales de comunicación asíncrona como pipes, conversaciones TCP, conversaciones Datakit, y lineas RS232 se implementan usando corrientes.
.PP
Una corriente comprende una lista lineal de 
.I "módulos en proceso" .
Cada módulo tiene una corriente de subida (hacia el proceso) y una de bajada (hacia el dispositivo)
.I "rutina put" .
Llamando a la rutina put del módulo en cualquiera de ambos lados se insertan datos en la corriente.
???????????
Cada módulo llama al siguiente para que envie datos hacia arriba o hacia abajo en la corriente.
Each module calls the succeeding one to send data up or down the stream.
??????????
.PP
Una instancia de un módulo en proceso se representa por un par de 
.I colas,
una para cada dirección.
Las colas apuntan a las rutinas put y pueden usarse para encolar información que va a pasarse por la corriente. Algunas rutinas put encolan los datos localmente y los envian a través de la corriente más tarde, bien sea por una llamada posterior o por un evento asíncrono como un temporizador de retransmisión o una interrupción de dispositivo.
Los módulos en proceso crean procesos ayudantes del nucleo que proporcionan un contexto para manejar los eventos asíncronos.Por ejemplo, un proceso ayudante del nucleo se despierta periodicamente para realizar cualquier retransmisión TCP necesaria.
????????
El uso de procesos del nucleo en lugar de rutinas de servicio que se ejecutan hasta que terminan difiere de la implementación de las corrientes en UNIX.
The use of kernel processes instead of serialized run-to-completion service routines
differs from the implementation of Unix streams.
????????
Las rutinas de servicio de UNIX no pueden usar ningún recurso bloqueante del nucleo y carecen de un estado local prolongado.
Los procesos ayudantes del nucleo resuelven estos problemas simplificando el código de las corrientes.
.PP
No hay sincronización implícita en nuestras corrientes. Cada módulo en proceso debe asegurarse de que los procesos concurrentes que usan la corriente están sincronizados.
Esto maximiza la concurrencia, pero introduce la posibilidad de un bloqueo mutuo.
Sin embargo, los bloqueos mutuos se evitan fácilmente mediante la programación cuidadosa; hasta la fecha no han causado problemas.
.PP
La información se representa por medio de listas enlazadas de estructuras del nucleo llamadas
.I bloques .
Cada bloque contiene un tipo, algunas banderas de estado, y punteros a un buffer opcional.
Los buffers de los bloques pueden guardar datos o información de control, es decir, directivas para los módulos en proceso.
Los bloques y los buffers de bloque se asignan dinámicamente de la memoria del nucleo.
.NH 3
Interfaz de usuario
.PP
Una corriente se representa a nivel de usuario por dos ficheros,
.CW ctl
y
.CW data .
Los nombres pueden ser cambiados por el driver de dispositivo que use la corriente, como ya dijimos en el ejemplo del driver de UART. El primer proceso que abre cualquiera de ellos crea la corriente automáticamente.
El último close la destruye.
Escribiendo en el fichero 
.CW data
????????
se copian datos en los bloques del nucleo y se pasan a la rutina put del primer módulo en proceso.
copies the data into kernel blocks and passes them to the downstream put routine of the first processing module.
?????????
Se garantiza que una escritura de menos de 32K cabe en un solo bloque. Las escrituras concurrentes en la misma corriente no están sincronizadas, aunque el tamaño de 34K asegura escrituras atómicas  para la mayoría de los protocolos.
El último bloque escrito es marcado con un delimitador para alertar a los módulos de la corriente que se ocupan de los límites de la escritura. La mayor parte de las veces, la primera rutina put llama a la segunda, la segunda a la tercera y así hasta que los datos salen. Como consecuencia, la mayoría de los datos salen sin necesidad de un cambio de contexto.
.PP
Leer del fichero
.CW data
devuelve los datos encolados en lo alto de la corriente. La lectura termina cuando se alcanza la cuenta de lecturas o cuando se encuentra el final de un bloque delimitado. Un bloqueo de lectura por corriente asegura que solamente un proceso puede leer de una corriente a la vez, y garantiza que los bytes leidos llegan contiguos desde la corriente. 
.PP
Al igual que las corrientes UNIX [Rit84a], las de Plan 9 pueden ser configuradas dinámicamente.
El sistema de corrientes intercepta e interpreta los siguientes bloques de control:
.IP "\f(CWpush\fP \fIname\fR" 15
agrega una instancia del modulo en proceso
.I name
en lo alto de la corriente.
.IP \f(CWpop\fP 15
retira el módulo superior de la stream.
.IP \f(CWhangup\fP 15
???????
envía un mensaje de colgar por la corriente desde el lado del dispositivo
sends a hangup message up the stream from the device end.
???????
.LP
Otros bloques de control son específicos de cada módulo y son interpretados por cada modulo en proceso sobre la marcha.
.PP
La enrevesada sintaxis y semántica de la llamada del sistema 
.CW ioctl
de UNIX nos convenció para eliminarla de Plan 9.
En realidad, 
.CW ioctl
es reemplazada por el fichero
.CW ctl.
Escribir en él
es idéntico a escribir en el fichero 
.CW data
solo que los bloques son de tipo
.I control .
Un módulo en proceso analiza cada bloque de control que ve. Los comandos en los bloques de control son cadenas ASCII, así que el orden de bytes no es un problema cuando un sistema controla corrientes en un espacio de nombres implementado en otro procesador. El tiempo para analizar bloques de control no es importante, puesto que las operaciones de control son raras.
.NH 3
Device Interface
.PP
The module at the downstream end of the stream is part of a device interface.
The particulars of the interface vary with the device.
Most device interfaces consist of an interrupt routine, an output
put routine, and a kernel process.
The output put routine stages data for the
device and starts the device if it is stopped.
The interrupt routine wakes up the kernel process whenever
the device has input to be processed or needs more output staged.
The kernel process puts information up the stream or stages more data for output.
The division of labor among the different pieces varies depending on
how much must be done at interrupt level.
However, the interrupt routine may not allocate blocks or call
a put routine since both actions require a process context.
.NH 3
Multiplexing
.PP
The conversations using a protocol device must be
multiplexed onto a single physical wire.
We push a multiplexer processing module
onto the physical device stream to group the conversations.
The device end modules on the conversations add the necessary header
onto downstream messages and then put them to the module downstream
of the multiplexer.
The multiplexing module looks at each message moving up its stream and
puts it to the correct conversation stream after stripping
the header controlling the demultiplexing.
.PP
This is similar to the Unix implementation of multiplexer streams.
The major difference is that we have no general structure that
corresponds to a multiplexer.
Each attempt to produce a generalized multiplexer created a more complicated
structure and underlined the basic difficulty of generalizing this mechanism.
We now code each multiplexer from scratch and favor simplicity over
generality.
.NH 3
Reflections
.PP
Despite five year's experience and the efforts of many programmers,
we remain dissatisfied with the stream mechanism.
Performance is not an issue;
the time to process protocols and drive
device interfaces continues to dwarf the
time spent allocating, freeing, and moving blocks
of data.
However the mechanism remains inordinately
complex.
Much of the complexity results from our efforts
to make streams dynamically configurable, to
reuse processing modules on different devices
and to provide kernel synchronization
to ensure data structures
don't disappear under foot.
This is particularly irritating since we seldom use these properties.
.PP
Streams remain in our kernel because we are unable to
devise a better alternative.
Larry Peterson's X-kernel [Pet89a]
is the closest contender but
doesn't offer enough advantage to switch.
If we were to rewrite the streams code, we would probably statically
allocate resources for a large fixed number of conversations and burn
memory in favor of less complexity.
.NH
The IL Protocol
.PP
None of the standard IP protocols is suitable for transmission of
9P messages over an Ethernet or the Internet.
TCP has a high overhead and does not preserve delimiters.
UDP, while cheap, does not provide reliable sequenced delivery.
Early versions of the system used a custom protocol that was
efficient but unsatisfactory for internetwork transmission.
When we implemented IP, TCP, and UDP we looked around for a suitable
replacement with the following properties:
.IP \(bu
Reliable datagram service with sequenced delivery
.IP \(bu
Runs over IP
.IP \(bu
Low complexity, high performance
.IP \(bu
Adaptive timeouts
.LP
None met our needs so a new protocol was designed.
IL is a lightweight protocol designed to be encapsulated by IP.
It is a connection-based protocol
providing reliable transmission of sequenced messages between machines.
No provision is made for flow control since the protocol is designed to transport RPC
messages between client and server.
A small outstanding message window prevents too
many incoming messages from being buffered;
messages outside the window are discarded
and must be retransmitted.
Connection setup uses a two way handshake to generate
initial sequence numbers at each end of the connection;
subsequent data messages increment the
sequence numbers allowing
the receiver to resequence out of order messages. 
In contrast to other protocols, IL does not do blind retransmission.
If a message is lost and a timeout occurs, a query message is sent.
The query message is a small control message containing the current
sequence numbers as seen by the sender.
The receiver responds to a query by retransmitting missing messages.
This allows the protocol to behave well in congested networks,
where blind retransmission would cause further
congestion.
Like TCP, IL has adaptive timeouts.
A round-trip timer is used
to calculate acknowledge and retransmission times in terms of the network speed.
This allows the protocol to perform well on both the Internet and on local Ethernets.
.PP
In keeping with the minimalist design of the rest of the kernel, IL is small.
The entire protocol is 847 lines of code, compared to 2200 lines for TCP.
IL is our protocol of choice.
.NH
Network Addressing
.PP
A uniform interface to protocols and devices is not sufficient to
support the transparency we require.
Since each network uses a different
addressing scheme,
the ASCII strings written to a control file have no common format.
As a result, every tool must know the specifics of the networks it
is capable of addressing.
Moreover, since each machine supplies a subset
of the available networks, each user must be aware of the networks supported
by every terminal and server machine.
This is obviously unacceptable.
.PP
Several possible solutions were considered and rejected; one deserves
more discussion.
We could have used a user-level file server
to represent the network name space as a Plan 9 file tree. 
This global naming scheme has been implemented in other distributed systems.
The file hierarchy provides paths to
directories representing network domains.
Each directory contains
files representing the names of the machines in that domain;
an example might be the path
.CW /net/name/usa/edu/mit/ai .
Each machine file contains information like the IP address of the machine.
We rejected this representation for several reasons.
First, it is hard to devise a hierarchy encompassing all representations
of the various network addressing schemes in a uniform manner.
Datakit and Ethernet address strings have nothing in common.
Second, the address of a machine is
often only a small part of the information required to connect to a service on
the machine.
For example, the IP protocols require symbolic service names to be mapped into
numeric port numbers, some of which are privileged and hence special.
Information of this sort is hard to represent in terms of file operations.
Finally, the size and number of the networks being represented burdens users with
an unacceptably large amount of information about the organization of the network
and its connectivity.
In this case the Plan 9 representation of a
resource as a file is not appropriate.
.PP
If tools are to be network independent, a third-party server must resolve
network names.
A server on each machine, with local knowledge, can select the best network
for any particular destination machine or service.
Since the network devices present a common interface,
the only operation which differs between networks is name resolution.
A symbolic name must be translated to
the path of the clone file of a protocol
device and an ASCII address string to write to the
.CW ctl
file.
A connection server (CS) provides this service.
.NH 2
Network Database
.PP
On most systems several
files such as
.CW /etc/hosts ,
.CW /etc/networks ,
.CW /etc/services ,
.CW /etc/hosts.equiv ,
.CW /etc/bootptab ,
and
.CW /etc/named.d
hold network information.
Much time and effort is spent
administering these files and keeping
them mutually consistent.
Tools attempt to
automatically derive one or more of the files from
information in other files but maintenance continues to be
difficult and error prone.
.PP
Since we were writing an entirely new system, we were free to
try a simpler approach.
One database on a shared server contains all the information
needed for network administration.
Two ASCII files comprise the main database:
.CW /lib/ndb/local
contains locally administered information and
.CW /lib/ndb/global
contains information imported from elsewhere.
The files contain sets of attribute/value pairs of the form
.I attr\f(CW=\fPvalue ,
where
.I attr
and
.I value
are alphanumeric strings.
Systems are described by multi-line entries;
a header line at the left margin begins each entry followed by zero or more
indented attribute/value pairs specifying
names, addresses, properties, etc.
For example, the entry for our CPU server
specifies a domain name, an IP address, an Ethernet address,
a Datakit address, a boot file, and supported protocols.
.P1
sys = helix
	dom=helix.research.bell-labs.com
	bootf=/mips/9power
	ip=135.104.9.31 ether=0800690222f0
	dk=nj/astro/helix
	proto=il flavor=9cpu
.P2
If several systems share entries such as
network mask and gateway, we specify that information
with the network or subnetwork instead of the system.
The following entries define a Class B IP network and 
a few subnets derived from it.
The entry for the network specifies the IP mask,
file system, and authentication server for all systems
on the network.
Each subnetwork specifies its default IP gateway.
.P1
ipnet=mh-astro-net ip=135.104.0.0 ipmask=255.255.255.0
	fs=bootes.research.bell-labs.com
	auth=1127auth
ipnet=unix-room ip=135.104.117.0
	ipgw=135.104.117.1
ipnet=third-floor ip=135.104.51.0
	ipgw=135.104.51.1
ipnet=fourth-floor ip=135.104.52.0
	ipgw=135.104.52.1
.P2
Database entries also define the mapping of service names
to port numbers for TCP, UDP, and IL.
.P1
tcp=echo	port=7
tcp=discard	port=9
tcp=systat	port=11
tcp=daytime	port=13
.P2
.PP
All programs read the database directly so
consistency problems are rare.
However the database files can become large.
Our global file, containing all information about
both Datakit and Internet systems in AT&T, has 43,000
lines.
To speed searches, we build hash table files for each
attribute we expect to search often.
The hash file entries point to entries
in the master files.
Every hash file contains the modification time of its master
file so we can avoid using an out-of-date hash table.
Searches for attributes that aren't hashed or whose hash table
is out-of-date still work, they just take longer.
.NH 2
Connection Server
.PP
On each system a user level connection server process, CS, translates
symbolic names to addresses.
CS uses information about available networks, the network database, and
other servers (such as DNS) to translate names.
CS is a file server serving a single file,
.CW /net/cs .
A client writes a symbolic name to
.CW /net/cs
then reads one line for each matching destination reachable
from this system.
The lines are of the form
.I "filename message",
where
.I filename
is the path of the clone file to open for a new connection and
.I message
is the string to write to it to make the connection.
The following example illustrates this.
.CW Ndb/csquery
is a program that prompts for strings to write to
.CW /net/cs
and prints the replies.
.P1
% ndb/csquery
> net!helix!9fs
/net/il/clone 135.104.9.31!17008
/net/dk/clone nj/astro/helix!9fs
.P2
.PP
CS provides meta-name translation to perform complicated
searches.
The special network name
.CW net
selects any network in common between source and
destination supporting the specified service.
A host name of the form \f(CW$\fIattr\f1
is the name of an attribute in the network database.
The database search returns the value
of the matching attribute/value pair
most closely associated with the source host.
Most closely associated is defined on a per network basis.
For example, the symbolic name
.CW tcp!$auth!rexauth
causes CS to search for the
.CW auth
attribute in the database entry for the source system, then its
subnetwork (if there is one) and then its network.
.P1
% ndb/csquery
> net!$auth!rexauth
/net/il/clone 135.104.9.34!17021
/net/dk/clone nj/astro/p9auth!rexauth
/net/il/clone 135.104.9.6!17021
/net/dk/clone nj/astro/musca!rexauth
.P2
.PP
Normally CS derives naming information from its database files.
For domain names however, CS first consults another user level
process, the domain name server (DNS).
If no DNS is reachable, CS relies on its own tables.
.PP
Like CS, the domain name server is a user level process providing
one file,
.CW /net/dns .
A client writes a request of the form
.I "domain-name type" ,
where
.I type
is a domain name service resource record type.
DNS performs a recursive query through the
Internet domain name system producing one line
per resource record found.  The client reads
.CW /net/dns 
to retrieve the records.
Like other domain name servers, DNS caches information
learned from the network.
DNS is implemented as a multi-process shared memory application
with separate processes listening for network and local requests.
.NH
Library routines
.PP
The section on protocol devices described the details
of making and receiving connections across a network.
The dance is straightforward but tedious.
Library routines are provided to relieve
the programmer of the details.
.NH 2
Connecting
.PP
The
.CW dial
library call establishes a connection to a remote destination.
It
returns an open file descriptor for the
.CW data
file in the connection directory.
.P1
int  dial(char *dest, char *local, char *dir, int *cfdp)
.P2
.IP \f(CWdest\fP 10
is the symbolic name/address of the destination.
.IP \f(CWlocal\fP 10
is the local address.
Since most networks do not support this, it is
usually zero.
.IP \f(CWdir\fP 10
is a pointer to a buffer to hold the path name of the protocol directory
representing this connection.
.CW Dial
fills this buffer if the pointer is non-zero.
.IP \f(CWcfdp\fP 10
is a pointer to a file descriptor for the
.CW ctl
file of the connection.
If the pointer is non-zero,
.CW dial
opens the control file and tucks the file descriptor here.
.LP
Most programs call
.CW dial
with a destination name and all other arguments zero.
.CW Dial
uses CS to
translate the symbolic name to all possible destination addresses
and attempts to connect to each in turn until one works.
Specifying the special name
.CW net
in the network portion of the destination
allows CS to pick a network/protocol in common
with the destination for which the requested service is valid.
For example, assume the system
.CW research.bell-labs.com
has the Datakit address
.CW nj/astro/research
and IP addresses
.CW 135.104.117.5
and
.CW 129.11.4.1 .
The call
.P1
fd = dial("net!research.bell-labs.com!login", 0, 0, 0, 0);
.P2
tries in succession to connect to
.CW nj/astro/research!login
on the Datakit and both
.CW 135.104.117.5!513
and
.CW 129.11.4.1!513
across the Internet.
.PP
.CW Dial
accepts addresses instead of symbolic names.
For example, the destinations
.CW tcp!135.104.117.5!513
and
.CW tcp!research.bell-labs.com!login
are equivalent
references to the same machine.
.NH 2
Listening
.PP
A program uses
four routines to listen for incoming connections.
It first
.CW announce() s
its intention to receive connections,
then
.CW listen() s
for calls and finally
.CW accept() s
or
.CW reject() s
them.
.CW Announce
returns an open file descriptor for the
.CW ctl
file of a connection and fills
.CW dir
with the
path of the protocol directory
for the announcement.
.P1
int  announce(char *addr, char *dir)
.P2
.CW Addr
is the symbolic name/address announced;
if it does not contain a service, the announcement is for
all services not explicitly announced.
Thus, one can easily write the equivalent of the
.CW inetd
program without
having to announce each separate service.
An announcement remains in force until the control file is
closed.
.LP
.CW Listen
returns an open file descriptor for the
.CW ctl
file and fills
.CW ldir
with the path
of the protocol directory
for the received connection.
It is passed
.CW dir
from the announcement.
.P1
int  listen(char *dir, char *ldir)
.P2
.LP
.CW Accept
and
.CW reject
are called with the control file descriptor and
.CW ldir
returned by
.CW listen.
Some networks such as Datakit accept a reason for a rejection;
networks such as IP ignore the third argument.
.P1
int  accept(int ctl, char *ldir)
int  reject(int ctl, char *ldir, char *reason)
.P2
.PP
The following code implements a typical TCP listener.
It announces itself, listens for connections, and forks a new
process for each.
The new process echoes data on the connection until the
remote end closes it.
The "*" in the symbolic name means the announcement is valid for
any addresses bound to the machine the program is run on.
.P1
.ta 8n 16n 24n 32n 40n 48n 56n 64n
int
echo_server(void)
{
	int dfd, lcfd;
	char adir[40], ldir[40];
	int n;
	char buf[256];

	afd = announce("tcp!*!echo", adir);
	if(afd < 0)
		return -1;

	for(;;){
		/* listen for a call */
		lcfd = listen(adir, ldir);
		if(lcfd < 0)
			return -1;

		/* fork a process to echo */
		switch(fork()){
		case 0:
			/* accept the call and open the data file */
			dfd = accept(lcfd, ldir);
			if(dfd < 0)
				return -1;

			/* echo until EOF */
			while((n = read(dfd, buf, sizeof(buf))) > 0)
				write(dfd, buf, n);
			exits(0);
		case -1:
			perror("forking");
		default:
			close(lcfd);
			break;
		}

	}
}
.P2
.NH
User Level
.PP
Communication between Plan 9 machines is done almost exclusively in
terms of 9P messages. Only the two services
.CW cpu
and
.CW exportfs
are used.
The
.CW cpu
service is analogous to
.CW rlogin .
However, rather than emulating a terminal session
across the network,
.CW cpu
creates a process on the remote machine whose name space is an analogue of the window
in which it was invoked.
.CW Exportfs
is a user level file server which allows a piece of name space to be
exported from machine to machine across a network. It is used by the
.CW cpu
command to serve the files in the terminal's name space when they are
accessed from the
cpu server.
.PP
By convention, the protocol and device driver file systems are mounted in a
directory called
.CW /net .
Although the per-process name space allows users to configure an
arbitrary view of the system, in practice their profiles build
a conventional name space.
.NH 2
Exportfs
.PP
.CW Exportfs
is invoked by an incoming network call.
The
.I listener
(the Plan 9 equivalent of
.CW inetd )
runs the profile of the user
requesting the service to construct a name space before starting
.CW exportfs .
After an initial protocol
establishes the root of the file tree being
exported,
the remote process mounts the connection,
allowing
.CW exportfs
to act as a relay file server. Operations in the imported file tree
are executed on the remote server and the results returned.
As a result
the name space of the remote machine appears to be exported into a
local file tree.
.PP
The
.CW import
command calls
.CW exportfs
on a remote machine, mounts the result in the local name space,
and
exits.
No local process is required to serve mounts;
9P messages are generated by the kernel's mount driver and sent
directly over the network.
.PP
.CW Exportfs
must be multithreaded since the system calls
.CW open,
.CW read
and
.CW write
may block.
Plan 9 does not implement the 
.CW select
system call but does allow processes to share file descriptors,
memory and other resources.
.CW Exportfs
and the configurable name space
provide a means of sharing resources between machines.
It is a building block for constructing complex name spaces
served from many machines.
.PP
The simplicity of the interfaces encourages naive users to exploit the potential
of a richly connected environment.
Using these tools it is easy to gateway between networks.
For example a terminal with only a Datakit connection can import from the server
.CW helix :
.P1
import -a helix /net
telnet ai.mit.edu
.P2
The
.CW import
command makes a Datakit connection to the machine
.CW helix
where
it starts an instance
.CW exportfs
to serve
.CW /net .
The
.CW import
command mounts the remote
.CW /net
directory after (the
.CW -a
option to
.CW import )
the existing contents
of the local
.CW /net
directory.
The directory contains the union of the local and remote contents of
.CW /net .
Local entries supersede remote ones of the same name so
networks on the local machine are chosen in preference
to those supplied remotely.
However, unique entries in the remote directory are now visible in the local
.CW /net 
directory.
All the networks connected to
.CW helix ,
not just Datakit,
are now available in the terminal. The effect on the name space is shown by the following
example:
.P1
philw-gnot% ls /net
/net/cs
/net/dk
philw-gnot% import -a musca /net
philw-gnot% ls /net
/net/cs
/net/cs
/net/dk
/net/dk
/net/dns
/net/ether
/net/il
/net/tcp
/net/udp
.P2
.NH 2
Ftpfs
.PP
We decided to make our interface to FTP
a file system rather than the traditional command.
Our command,
.I ftpfs,
dials the FTP port of a remote system, prompts for login and password, sets image mode,
and mounts the remote file system onto
.CW /n/ftp .
Files and directories are cached to reduce traffic.
The cache is updated whenever a file is created.
Ftpfs works with TOPS-20, VMS, and various Unix flavors
as the remote system.
.NH
Cyclone Fiber Links
.PP
The file servers and CPU servers are connected by
high-bandwidth
point-to-point links.
A link consists of two VME cards connected by a pair of optical
fibers.
The VME cards use 33MHz Intel 960 processors and AMD's TAXI
fiber transmitter/receivers to drive the lines at 125 Mbit/sec.
Software in the VME card reduces latency by copying messages from system memory
to fiber without intermediate buffering.
.NH
Performance
.PP
We measured both latency and throughput
of reading and writing bytes between two processes
for a number of different paths.
Measurements were made on two- and four-CPU SGI Power Series processors.
The CPUs are 25 MHz MIPS 3000s.
The latency is measured as the round trip time
for a byte sent from one process to another and
back again.
Throughput is measured using 16k writes from
one process to another.
.DS C
.TS
box, tab(:);
c s s
c | c | c
l | n | n.
Table 1 - Performance
_
test:throughput:latency
:MBytes/sec:millisec
_
pipes:8.15:.255
_
IL/ether:1.02:1.42
_
URP/Datakit:0.22:1.75
_
Cyclone:3.2:0.375
.TE
.DE
.NH
Conclusion
.PP
The representation of all resources as file systems
coupled with an ASCII interface has proved more powerful
than we had originally imagined.
Resources can be used by any computer in our networks
independent of byte ordering or CPU type.
The connection server provides an elegant means
of decoupling tools from the networks they use.
Users successfully use Plan 9 without knowing the
topology of the system or the networks they use.
More information about 9P can be found in the Section 5 of the Plan 9 Programmer's
Manual, Volume I.
.NH
References
.LP
[Pike90] R. Pike, D. Presotto, K. Thompson, H. Trickey,
``Plan 9 from Bell Labs'',
.I
UKUUG Proc. of the Summer 1990 Conf. ,
London, England,
1990.
.LP
[Needham] R. Needham, ``Names'', in
.I
Distributed systems,
.R
S. Mullender, ed.,
Addison Wesley, 1989.
.LP
[Presotto] D. Presotto, ``Multiprocessor Streams for Plan 9'',
.I
UKUUG Proc. of the Summer 1990 Conf. ,
.R
London, England, 1990.
.LP
[Met80] R. Metcalfe, D. Boggs, C. Crane, E. Taf and J. Hupp, ``The
Ethernet Local Network: Three reports'',
.I
CSL-80-2,
.R
XEROX Palo Alto Research Center, February 1980.
.LP
[Fra80] A. G. Fraser, ``Datakit - A Modular Network for Synchronous
and Asynchronous Traffic'', 
.I
Proc. Int'l Conf. on Communication,
.R
Boston, June 1980.
.LP
[Pet89a] L. Peterson, ``RPC in the X-Kernel: Evaluating new Design Techniques'',
.I
Proc. Twelfth Symp. on Op. Sys. Princ.,
.R
Litchfield Park, AZ, December 1990.
.LP
[Rit84a] D. M. Ritchie, ``A Stream Input-Output System'',
.I
AT&T Bell Laboratories Technical Journal, 68(8),
.R
October 1984.
